from ibm_ray_config.modules.config_builder import ConfigBuilder, update_decorator, spinner
from typing import Any, Dict
from ibm_ray_config.modules.utils import Color, color_msg, find_obj, find_default


class ImageConfig(ConfigBuilder):
    
    def __init__(self, base_config: Dict[str, Any]) -> None:
        super().__init__(base_config)

    @spinner
    def get_image_objects(self):
        """returns a list of images with amd architecture, sorted by name.  

        uses a paginator to collect images beyond page limit amount.
        """
        images = []
        res = self.ibm_vpc_client.list_images().get_result()
        filtered_res  = filter_images(res['images'],['operating_system','architecture'],"amd" )
        images.extend(filtered_res)
        while res.get('next',None):
            link_to_next = res['next']['href'].split('start=')[1].split('&limit')[0]
            res = self.ibm_vpc_client.list_images(start=link_to_next).get_result()
            filtered_res  = filter_images(res['images'],['operating_system','architecture'],"amd" )
            images.extend(filtered_res)
        
        return sorted(images,key = lambda img:img['name'])

    @update_decorator
    def run(self) -> Dict[str, Any]:

        image_objects = self.get_image_objects()

        default = find_default({'name': 'ibm-ubuntu-20-04-'}, image_objects, name='name', substring=True)
        image_obj = find_obj(image_objects, 'Please choose an image. Ubuntu image is advised as node setup is using apt.', default=default)
        
        if 'ubuntu' not in image_obj['name']: 
            print(color_msg("Node setup commands are suited to Ubuntu.\nAlter the cluster config setup commands.",Color.RED))

        return image_obj['id'], image_obj['minimum_provisioned_size'], image_obj['owner_type'] == 'user'

    @update_decorator
    def verify(self, base_config):
        image_id = self.defaults['image_id']
        image_objects = self.ibm_vpc_client.list_images().get_result()['images']
        if image_id:
            image_obj = find_obj(image_objects, 'dummy', obj_id=image_id)
        else:
            # find first occurrence
            image_obj = next((obj for obj in image_objects if 'ibm-ubuntu-20-04-' in obj['name']), None)
            
        return image_obj['id'], image_obj['minimum_provisioned_size'], image_obj['owner_type'] == 'user'

    @update_decorator
    def create_default(self):
        image_objects = self.ibm_vpc_client.list_images().get_result()['images']

        image_obj = next((image for image in image_objects if 'ibm-ubuntu-20-04-' in image['name']), None)
        
        print(f'Selected \033[92mUbuntu\033[0m 20.04 VM image, {image_obj["name"]}')
        return image_obj['id'], image_obj['minimum_provisioned_size'], image_obj['owner_type'] == 'user'

def filter_images(images, fields, value):
    """returns filtered list of images matching the prefix of the specified value of a given field
    
    images (dict) 
    fields (str) - list of dict possibly multilevel dict fields 
    value (str) - prefix of required value   
    """
    filtered_images = []
    for image in images:
        img_copy = image.copy()
        for field in fields:
            img_copy = img_copy.get(field)
        if img_copy and img_copy.startswith(value): 
            filtered_images.append(image) 
    return filtered_images


import re
import copy
from typing import Any, Dict

import inquirer
from ibm_ray_config.modules.config_builder import ConfigBuilder
from ibm_ray_config.modules.utils import validate_cluster_name, get_profile_resources

class WorkersConfig(ConfigBuilder):

    def run(self) -> Dict[str, Any]:
        default_cluster_name = self.base_config.get('cluster_name', 'default')
        default_min_workers = self.base_config.get('min_workers', '0')
        default_max_workers = default_min_workers

        question = [
            inquirer.Text(
                'name', message="Enter cluster name following the pattern `[a-z]|[a-z][-a-z0-9]*[a-z0-9]`", validate = validate_cluster_name, default=default_cluster_name),
            inquirer.Text('min_workers', message="Minimum number of worker nodes",
                          default=default_min_workers, validate=lambda _, x: re.match('^[+]?[0-9]+$', x)),
            inquirer.Text('max_workers', message="Maximum number of worker nodes", default=default_max_workers,
                          validate=lambda answers, x: re.match('^[+]?[0-9]+$', x) and int(x) >= int(answers['min_workers']))
        ]
        answers = inquirer.prompt(question, raise_keyboard_interrupt=True)
        self.base_config['cluster_name'] = answers['name']
        self.base_config['max_workers'] = int(answers['max_workers'])

        if self.base_config.get('worker_instance_profile', None):
            self.base_config['available_node_types']['ray_head_default']['min_workers'] = 0 
            self.base_config['available_node_types']['ray_head_default']['max_workers'] = 0

            worker_dict = copy.deepcopy(self.base_config['available_node_types']['ray_head_default'])
            worker_dict['node_config'].pop('head_ip',None)

            worker_dict['min_workers'] = int(answers['min_workers'])
            worker_dict['max_workers'] = int(answers['max_workers'])
            worker_dict['node_config']['instance_profile_name'] = self.base_config['worker_instance_profile']
            
            cpu, memory, gpu = get_profile_resources(self.base_config['worker_instance_profile'])
            if gpu:
                worker_dict['resources']['GPU'] = gpu
            worker_dict['resources']['CPU'] = cpu
            worker_dict['resources']['memory'] = memory

            self.base_config['available_node_types']['ray_worker_default'] = worker_dict
            del self.base_config['worker_instance_profile']
        else:
            self.base_config['available_node_types']['ray_head_default']['min_workers'] = int(answers['min_workers'])
            self.base_config['available_node_types']['ray_head_default']['max_workers'] = int(answers['max_workers'])

        return self.base_config
    
    def verify(self, base_config):
        min_workers = base_config['available_node_types']['ray_head_default']['min_workers']
        max_workers = base_config['available_node_types']['ray_head_default']['max_workers']
        
        if max_workers < min_workers:
            raise Exception(f'specified min workers {min_workers} larger than max workers {max_workers}')

        return base_config
